// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file ProblemImplicit.h */

#ifndef AMDIS_PROBLEM_IMPLICIT_H
#define AMDIS_PROBLEM_IMPLICIT_H

#include "ProblemStat.h"

namespace AMDiS {
  
  using namespace std;

  class ProblemImplicitBase 
  {
  public:
    virtual ~ProblemImplicitBase() {}

    virtual bool createImplicitMesh() = 0;

    virtual DOFVector<double>* getSignedDistance(unsigned int im = 0, 
						 unsigned int m = 0) = 0;

    virtual DOFVector<double>* getPhi1(unsigned int im = 0, 
				       unsigned int m = 0) = 0;

    virtual DOFVector<double>* getPhi2(unsigned int im = 0, 
				       unsigned int m = 0) = 0;

    virtual DOFVector<double>* getLevelset(unsigned int im = 0, 
					   unsigned int m = 0) = 0;

  protected:
    void readDofVec(istream& , DOFVector<double>* , Mesh*);

    void readR(istream&, double, Mesh*, int implMesh = 0, int comp = 0);

    void readPhi1(istream&, double, Mesh*, int implMesh = 0, int comp = 0);

    void readPhi2(istream&, double, Mesh*, int implMesh = 0, int comp = 0);

    string getDofFilename(string path, int implMesh = 0);

    double getEpsilon(string path, int implMesh = 0);

    int getType(string path, int implMesh = 0);

    virtual bool isImplicitMesh(int comp = 0) = 0;
  };
 

  class ProblemImplicit : public ProblemStatSeq, 
			  public ProblemImplicitBase
  {
  public:
    ProblemImplicit(string name, 
		    ProblemIterationInterface* problem = NULL)
      : ProblemStatSeq(name, problem),
	r(0),
	phi1(0),
	phi2(0),
	levelSet(0),
        implMesh(0)
    {}

    virtual void createMesh();

    virtual bool createImplicitMesh();

    virtual void initialize(Flag initFlag, 
			    ProblemStatSeq* adoptProblem = NULL, 
		            Flag adoptFlag = INIT_NOTHING);

    DOFVector<double>* getSignedDistance(unsigned int im = 0, unsigned int m = 0);

    DOFVector<double>* getPhi1(unsigned int im = 0, unsigned int m = 0); 

    DOFVector<double>* getPhi2(unsigned int im = 0, unsigned int m = 0);

    DOFVector<double>* getLevelset(unsigned int im = 0, unsigned int m = 0);
 
  protected:
    bool createImplicitMesh(int p);

    bool createImplicitMesh(string, int, int); 

    /// DOFVector for a signed distance
    vector< vector<DOFVector<double>*> > r;

    /// DOFVector for the phasefield function 0.5*(1-tanh(3*r/eps))
    vector< vector<DOFVector<double>*> > phi1;

    /// DOFVector for the phasefield function 0.5*(1+tanh(3*r/eps))
    vector< vector<DOFVector<double>*> > phi2;

    /// DOFVector for the levelset function 
    /// (levelSet(x): x \in \Omega: 1, x \not \in Omega: -1, x \in \Gamma: 0)
    vector< vector<DOFVector<double>*> >  levelSet;

    vector<bool> implMesh;
    
    bool isImplicitMesh(int comp = 0) 
    { 
      return implMesh[comp]; 
    }
  };
}
#endif
