// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file MeshManipulation.h */

#ifndef AMDIS_MESHMANIPULATION_H
#define AMDIS_MESHMANIPULATION_H

#include <set>
#include <map>

#include "AMDiS_fwd.h"
#include "parallel/ElementObjectData.h"

namespace AMDiS {

  class MeshManipulation {
  public:
    MeshManipulation(FiniteElemSpace *space);
    
    ~MeshManipulation();

    void deleteDoubleDofs(std::set<MacroElement*>& newMacroEl, ElementObjects &elObj);

    /** \brief
     * Starts the procedure to fit a given edge/face of an element with a mesh
     * structure code. This functions prepares some data structures and call
     * then \ref fitElementToMeshCode, that mainly refines the element such that
     * it fits to the mesh structure code.
     *
     * \param[in] code      The mesh structure code to which the edge/face of
     *                      an element must be fitted.
     * \param[in] boundEl   Defines the element and its sub-object, i.e., edge or
     *                      face that must be fitted.
     *
     * \return   Returns true, if the mesh was changed, i.e. refined, to fit the
     *           element to the structure code.
     */
    bool fitElementToMeshCode(MeshStructure &code, BoundaryObject &boundEl);

  private:
    /** \brief
     * Recursively fits a given mesh structure code to an edge/face of an element.
     * This function is always initialy called from \ref startFitElementToMeshCode.
     *
     * \param[in] code         The mesh structure code which is used to fit an
     *                         edge/face of an element.
     * \param[in] stack        A traverse stack object. The upper most element in this
     *                         stack must be used for fitting the mesh structure code
     *                         at the current position.
     * \param[in] subObj       Defines whether an edge or a face must be fitted.
     * \param[in] ithObj       Defines which edge/face must be fitted.
     * \param[in] reverseMode  Defines, whether the mesh structure code is given
     *                         in reverse mode, i.e., left and right children where
     *                         changed when the code was created.
     *
     * \return   Returns true, if the mesh was changed, i.e. refined, to fit the
     *           element to the structure code.
     */
    bool fitElementToMeshCode(MeshStructure &code, 
			      TraverseStack &stack,
			      GeoIndex subObj,
			      int ithObj,
			      bool reverseMode);


    
  private:
    FiniteElemSpace *feSpace;
	     
    Mesh *mesh;

    RefinementManager *refineManager;
  };

}

#endif

