#include "HeatProject.h"
#include "TimedObject.h"

/// Dirichlet boundary function
class G : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  
  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    return sin(M_PI * (*timePtr)) * exp(-10.0 * (x * x));
  }
};

/// RHS function
class F : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {}

  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    int dim = x.getSize();
    double r2 = x * x;
    double ux = sin(M_PI * (*timePtr)) * exp(-10.0 * r2);
    double ut = M_PI * cos(M_PI * (*timePtr)) * exp(-10.0 * r2);
    return ut -(400.0 * r2 - 20.0 * dim) * ux;
  }
};

Heatdemo::Heatdemo():
  heatSpace("heat->space"),
  heat(heatSpace),
  adaptInfo(NULL),
  adaptInfoInitial(NULL),
  A(NULL),
  C(NULL),
  Fop(NULL)
{}

Heatdemo::~Heatdemo() {
  if(adaptInfo != NULL)
    delete adaptInfo;
  if(adaptInfoInitial != NULL)
    delete adaptInfoInitial;

  if(A != NULL)
    delete A;
  if(C != NULL)
    delete C;
  if(Fop != NULL)
    delete Fop;
}

void Heatdemo::create(string& filename) {
    // ===== init parameters =====
  Parameters::init(false, filename);
//  Parameters::readArgv(argc, argv);


  // ===== create and init stationary problem =====
  heatSpace.initialize(INIT_ALL);


  // ===== create instationary problem =====
  heat.initialize(INIT_ALL);

  // create adapt info
  adaptInfo = new AdaptInfo("heat->adapt");

  // create initial adapt info
  adaptInfoInitial = new AdaptInfo("heat->initial->adapt");

  // create instationary adapt
  adaptInstat = new AdaptInstationary("heat->adapt",
				&heatSpace,
				adaptInfo,
				&heat,
				adaptInfoInitial);


  // ===== create rhs functions =====
  int degree = heatSpace.getFeSpace()->getBasisFcts()->getDegree();
  F *rhsFct = new F(degree);
  rhsFct->setTimePtr(heat.getRHSTimePtr());


  // ===== create operators =====
  double one = 1.0;
  double zero = 0.0;

  // create laplace
  A = new Operator(heatSpace.getFeSpace());
  A->addSecondOrderTerm(new Laplace_SOT);
  A->setUhOld(heat.getOldSolution());
  if (*(heat.getThetaPtr()) != 0.0)
    heatSpace.addMatrixOperator(A, heat.getThetaPtr(), &one);
  if (*(heat.getTheta1Ptr()) != 0.0)
    heatSpace.addVectorOperator(A, heat.getTheta1Ptr(), &zero);

  // create zero order operator
  C = new Operator(heatSpace.getFeSpace());
  C->addZeroOrderTerm(new Simple_ZOT);
  C->setUhOld(heat.getOldSolution());
  heatSpace.addMatrixOperator(C, heat.getTau1Ptr(), heat.getTau1Ptr());
  heatSpace.addVectorOperator(C, heat.getTau1Ptr(), heat.getTau1Ptr());

  // create RHS operator
  Fop = new Operator(heatSpace.getFeSpace());
  Fop->addZeroOrderTerm(new CoordsAtQP_ZOT(rhsFct));
  heatSpace.addVectorOperator(Fop);


  // ===== create boundary functions =====
  G *boundaryFct = new G;
  boundaryFct->setTimePtr(heat.getBoundaryTimePtr());
  heat.setExactSolution(boundaryFct);

}

int Heatdemo::solve(SolutionInformation& info) {
  int retCode = adaptInstat->adapt();
  info.dofVec = heatSpace.getSolution();
  return retCode;
}
